from rdflib import Graph

from rdflib.query import Result, ResultParser


class GraphResultParser(ResultParser):
    def parse(self, source, content_type):

        res = Result("CONSTRUCT")  # hmm - or describe?type_)
        res.graph = Graph()
        res.graph.parse(source, format=content_type)

        return res



"""
This implements the Tab Separated SPARQL Result Format

It is implemented with pyparsing, reusing the elements from the SPARQL Parser
"""

import codecs

from pyparsing import (
    Optional,
    ZeroOrMore,
    Literal,
    ParserElement,
    ParseException,
    Suppress,
    FollowedBy,
    LineEnd,
)

from rdflib.query import Result, ResultParser

from rdflib.plugins.sparql.parser import (
    Var,
    STRING_LITERAL1,
    STRING_LITERAL2,
    IRIREF,
    BLANK_NODE_LABEL,
    NumericLiteral,
    BooleanLiteral,
    LANGTAG,
)
from rdflib.plugins.sparql.parserutils import Comp, Param, CompValue

from rdflib import Literal as RDFLiteral

ParserElement.setDefaultWhitespaceChars(" \n")


String = STRING_LITERAL1 | STRING_LITERAL2

RDFLITERAL = Comp(
    "literal",
    Param("string", String)
    + Optional(
        Param("lang", LANGTAG.leaveWhitespace())
        | Literal("^^").leaveWhitespace() + Param("datatype", IRIREF).leaveWhitespace()
    ),
)

NONE_VALUE = object()

EMPTY = FollowedBy(LineEnd()) | FollowedBy("\t")
EMPTY.setParseAction(lambda x: NONE_VALUE)

TERM = RDFLITERAL | IRIREF | BLANK_NODE_LABEL | NumericLiteral | BooleanLiteral

ROW = (EMPTY | TERM) + ZeroOrMore(Suppress("\t") + (EMPTY | TERM))
ROW.parseWithTabs()

HEADER = Var + ZeroOrMore(Suppress("\t") + Var)
HEADER.parseWithTabs()


class TSVResultParser(ResultParser):
    def parse(self, source, content_type=None):

        if isinstance(source.read(0), bytes):
            # if reading from source returns bytes do utf-8 decoding
            source = codecs.getreader("utf-8")(source)

        try:
            r = Result("SELECT")

            header = source.readline()

            r.vars = list(HEADER.parseString(header.strip(), parseAll=True))
            r.bindings = []
            while True:
                line = source.readline()
                if not line:
                    break
                line = line.strip("\n")
                if line == "":
                    continue

                row = ROW.parseString(line, parseAll=True)
                r.bindings.append(dict(zip(r.vars, (self.convertTerm(x) for x in row))))

            return r

        except ParseException as err:
            print(err.line)
            print(" " * (err.column - 1) + "^")
            print(err)

    def convertTerm(self, t):
        if t is NONE_VALUE:
            return None
        if isinstance(t, CompValue):
            if t.name == "literal":
                return RDFLiteral(t.string, lang=t.lang, datatype=t.datatype)
            else:
                raise Exception("I dont know how to handle this: %s" % (t,))
        else:
            return t


if __name__ == "__main__":
    import sys

    r = Result.parse(source=sys.argv[1], format="tsv")
    print(r.vars)
    print(r.bindings)
    # print r.serialize(format='json')



"""
Parser plugin interface.

This module defines the parser plugin interface and contains other
related parser support code.

The module is mainly useful for those wanting to write a parser that
can plugin to rdflib. If you are wanting to invoke a parser you likely
want to do so through the Graph class parse method.

"""

import codecs
import os
import pathlib
import sys

from io import BytesIO, TextIOBase, TextIOWrapper, StringIO, BufferedIOBase

from urllib.request import Request
from urllib.request import url2pathname
from urllib.parse import urljoin
from urllib.request import urlopen
from urllib.error import HTTPError

from xml.sax import xmlreader

from rdflib import __version__
from rdflib.term import URIRef
from rdflib.namespace import Namespace

__all__ = [
    "Parser",
    "InputSource",
    "StringInputSource",
    "URLInputSource",
    "FileInputSource",
]


class Parser(object):
    __slots__ = ()

    def __init__(self):
        pass

    def parse(self, source, sink):
        pass


class BytesIOWrapper(BufferedIOBase):
    __slots__ = ("wrapped", "encoded", "encoding")

    def __init__(self, wrapped: str, encoding="utf-8"):
        super(BytesIOWrapper, self).__init__()
        self.wrapped = wrapped
        self.encoding = encoding
        self.encoded = None

    def read(self, *args, **kwargs):
        if self.encoded is None:
            b, blen = codecs.getencoder(self.encoding)(self.wrapped)
            self.encoded = BytesIO(b)
        return self.encoded.read(*args, **kwargs)

    def read1(self, *args, **kwargs):
        if self.encoded is None:
            b = codecs.getencoder(self.encoding)(self.wrapped)
            self.encoded = BytesIO(b)
        return self.encoded.read1(*args, **kwargs)

    def readinto(self, *args, **kwargs):
        raise NotImplementedError()

    def readinto1(self, *args, **kwargs):
        raise NotImplementedError()

    def write(self, *args, **kwargs):
        raise NotImplementedError()


class InputSource(xmlreader.InputSource, object):
    """
    TODO:
    """

    def __init__(self, system_id=None):
        xmlreader.InputSource.__init__(self, system_id=system_id)
        self.content_type = None
        self.auto_close = False  # see Graph.parse(), true if opened by us

    def close(self):
        c = self.getCharacterStream()
        if c and hasattr(c, "close"):
            try:
                c.close()
            except Exception:
                pass
        f = self.getByteStream()
        if f and hasattr(f, "close"):
            try:
                f.close()
            except Exception:
                pass


class StringInputSource(InputSource):
    """
    Constructs an RDFLib Parser InputSource from a Python String or Bytes
    """

    def __init__(self, value, encoding="utf-8", system_id=None):
        super(StringInputSource, self).__init__(system_id)
        if isinstance(value, str):
            stream = StringIO(value)
            self.setCharacterStream(stream)
            self.setEncoding(encoding)
            b_stream = BytesIOWrapper(value, encoding)
            self.setByteStream(b_stream)
        else:
            stream = BytesIO(value)
            self.setByteStream(stream)
            c_stream = TextIOWrapper(stream, encoding)
            self.setCharacterStream(c_stream)
            self.setEncoding(c_stream.encoding)


headers = {
    "User-agent": "rdflib-%s (http://rdflib.net/; eikeon@eikeon.com)" % __version__
}


class URLInputSource(InputSource):
    """
    TODO:
    """

    def __init__(self, system_id=None, format=None):
        super(URLInputSource, self).__init__(system_id)
        self.url = system_id

        # copy headers to change
        myheaders = dict(headers)
        if format == "application/rdf+xml":
            myheaders["Accept"] = "application/rdf+xml, */*;q=0.1"
        elif format == "n3":
            myheaders["Accept"] = "text/n3, */*;q=0.1"
        elif format == "turtle":
            myheaders["Accept"] = "text/turtle,application/x-turtle, */*;q=0.1"
        elif format == "nt":
            myheaders["Accept"] = "text/plain, */*;q=0.1"
        elif format == "json-ld":
            myheaders[
                "Accept"
            ] = "application/ld+json, application/json;q=0.9, */*;q=0.1"
        else:
            myheaders["Accept"] = (
                "application/rdf+xml,text/rdf+n3;q=0.9,"
                + "application/xhtml+xml;q=0.5, */*;q=0.1"
            )

        req = Request(system_id, None, myheaders)

        def _urlopen(req: Request):
            try:
                return urlopen(req)
            except HTTPError as ex:
                # 308 (Permanent Redirect) is not supported by current python version(s)
                # See https://bugs.python.org/issue40321
                # This custom error handling should be removed once all
                # supported versions of python support 308.
                if ex.code == 308:
                    req.full_url = ex.headers.get("Location")
                    return _urlopen(req)
                else:
                    raise

        file = _urlopen(req)
        # Fix for issue 130 https://github.com/RDFLib/rdflib/issues/130
        self.url = file.geturl()  # in case redirections took place
        self.setPublicId(self.url)
        self.content_type = file.info().get("content-type")
        if self.content_type is not None:
            self.content_type = self.content_type.split(";", 1)[0]
        self.setByteStream(file)
        # TODO: self.setEncoding(encoding)
        self.response_info = file.info()  # a mimetools.Message instance

    def __repr__(self):
        return self.url


class FileInputSource(InputSource):
    def __init__(self, file):
        base = pathlib.Path.cwd().as_uri()
        system_id = URIRef(pathlib.Path(file.name).absolute().as_uri(), base=base)
        super(FileInputSource, self).__init__(system_id)
        self.file = file
        if isinstance(file, TextIOBase):  # Python3 unicode fp
            self.setCharacterStream(file)
            self.setEncoding(file.encoding)
            try:
                b = file.buffer
                self.setByteStream(b)
            except (AttributeError, LookupError):
                self.setByteStream(file)
        else:
            self.setByteStream(file)
            # We cannot set characterStream here because
            # we do not know the Raw Bytes File encoding.

    def __repr__(self):
        return repr(self.file)


def create_input_source(
    source=None, publicID=None, location=None, file=None, data=None, format=None
):
    """
    Return an appropriate InputSource instance for the given
    parameters.
    """

    # test that exactly one of source, location, file, and data is not None.
    non_empty_arguments = list(
        filter(
            lambda v: v is not None,
            [source, location, file, data],
        )
    )

    if len(non_empty_arguments) != 1:
        raise ValueError(
            "exactly one of source, location, file or data must be given",
        )

    input_source = None

    if source is not None:
        if isinstance(source, InputSource):
            input_source = source
        else:
            if isinstance(source, str):
                location = source
            elif isinstance(source, pathlib.PurePath):
                location = str(source)
            elif isinstance(source, bytes):
                data = source
            elif hasattr(source, "read") and not isinstance(source, Namespace):
                f = source
                input_source = InputSource()
                if hasattr(source, "encoding"):
                    input_source.setCharacterStream(source)
                    input_source.setEncoding(source.encoding)
                    try:
                        b = file.buffer
                        input_source.setByteStream(b)
                    except (AttributeError, LookupError):
                        input_source.setByteStream(source)
                else:
                    input_source.setByteStream(f)
                if f is sys.stdin:
                    input_source.setSystemId("file:///dev/stdin")
                elif hasattr(f, "name"):
                    input_source.setSystemId(f.name)
            else:
                raise Exception(
                    "Unexpected type '%s' for source '%s'" % (type(source), source)
                )

    absolute_location = None  # Further to fix for issue 130

    auto_close = False  # make sure we close all file handles we open

    if location is not None:
        (
            absolute_location,
            auto_close,
            file,
            input_source,
        ) = _create_input_source_from_location(
            file=file,
            format=format,
            input_source=input_source,
            location=location,
        )

    if file is not None:
        input_source = FileInputSource(file)

    if data is not None:
        if not isinstance(data, (str, bytes, bytearray)):
            raise RuntimeError("parse data can only str, or bytes.")
        input_source = StringInputSource(data)
        auto_close = True

    if input_source is None:
        raise Exception("could not create InputSource")
    else:
        input_source.auto_close |= auto_close
        if publicID is not None:  # Further to fix for issue 130
            input_source.setPublicId(publicID)
        # Further to fix for issue 130
        elif input_source.getPublicId() is None:
            input_source.setPublicId(absolute_location or "")
        return input_source


def _create_input_source_from_location(file, format, input_source, location):
    # Fix for Windows problem https://github.com/RDFLib/rdflib/issues/145
    path = pathlib.Path(location)
    if path.exists():
        location = path.absolute().as_uri()

    base = pathlib.Path.cwd().as_uri()

    absolute_location = URIRef(location, base=base)

    if absolute_location.startswith("file:///"):
        filename = url2pathname(absolute_location.replace("file:///", "/"))
        file = open(filename, "rb")
    else:
        input_source = URLInputSource(absolute_location, format)

    auto_close = True
    # publicID = publicID or absolute_location  # Further to fix
    # for issue 130

    return absolute_location, auto_close, file, input_source



from rdflib.term import URIRef
from rdflib.namespace import DefinedNamespace, Namespace


class XSD(DefinedNamespace):
    """
    W3C XML Schema Definition Language (XSD) 1.1 Part 2: Datatypes

    Generated from: ../schemas/datatypes.xsd
    Date: 2020-05-26 14:21:14.993677

    """

    ENTITIES: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#ENTITIES
    ENTITY: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#ENTITY
    ID: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#ID
    IDREF: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#IDREF
    IDREFS: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#IDREFS
    NCName: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#NCName
    NMTOKEN: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#NMTOKEN
    NMTOKENS: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#NMTOKENS
    NOTATION: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#NOTATIONNOTATION cannot be used directly in a schema; rather a type
    Name: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#Name
    QName: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#QName
    anyURI: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#anyURI
    base64Binary: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#base64Binary
    boolean: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#boolean
    byte: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#byte
    date: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#date
    dateTime: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#dateTime
    dayTimeDuration: URIRef  # see: http://www.w3.org/TR/xmlschema11-2/#dayTimeDuration
    decimal: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#decimal
    double: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#double
    duration: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#duration
    float: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#float
    gDay: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#gDay
    gMonth: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#gMonth
    gMonthDay: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#gMonthDay
    gYear: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#gYear
    gYearMonth: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#gYearMonth
    hexBinary: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#binary
    int: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#int
    integer: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#integer
    language: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#language
    long: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#long
    negativeInteger: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#negativeInteger
    nonNegativeInteger: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#nonNegativeInteger
    nonPositiveInteger: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#nonPositiveInteger
    normalizedString: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#normalizedString
    positiveInteger: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#positiveInteger
    short: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#short
    string: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#string
    time: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#time
    token: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#token
    unsignedByte: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#unsignedByte
    unsignedInt: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#unsignedInt
    unsignedLong: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#unsignedLong
    unsignedShort: URIRef  # see: http://www.w3.org/TR/xmlschema-2/#unsignedShort
    yearMonthDuration: URIRef  # see: http://www.w3.org/TR/xmlschema11-2/#yearMonthDuration

    _NS = Namespace("http://www.w3.org/2001/XMLSchema#")



from rdflib.term import URIRef
from rdflib.namespace import DefinedNamespace, Namespace


class DOAP(DefinedNamespace):
    """
    Description of a Project (DOAP) vocabulary

    The Description of a Project (DOAP) vocabulary, described using W3C RDF Schema and the Web Ontology Language.

    Generated from: http://usefulinc.com/ns/doap
    Date: 2020-05-26 14:20:01.307972

    """

    _fail = True

    # http://www.w3.org/1999/02/22-rdf-syntax-ns#Property
    audience: URIRef  # Description of target user base
    blog: URIRef  # URI of a blog related to a project
    browse: URIRef  # Web browser interface to repository.
    category: URIRef  # A category of project.
    created: URIRef  # Date when something was created, in YYYY-MM-DD form. e.g. 2004-04-05
    description: URIRef  # Plain text description of a project, of 2-4 sentences in length.
    developer: URIRef  # Developer of software for the project.
    documenter: URIRef  # Contributor of documentation to the project.
    helper: URIRef  # Project contributor.
    implements: URIRef  # A specification that a project implements. Could be a standard, API or legally defined level of conformance.
    language: URIRef  # ISO language code a project has been translated into
    license: URIRef  # The URI of an RDF description of the license the software is distributed under. E.g. a SPDX reference
    location: URIRef  # Location of a repository.
    maintainer: URIRef  # Maintainer of a project, a project leader.
    module: URIRef  # Module name of a Subversion, CVS, BitKeeper or Arch repository.
    name: URIRef  # A name of something.
    os: URIRef  # Operating system that a project is limited to.  Omit this property if the project is not OS-specific.
    platform: URIRef  # Indicator of software platform (non-OS specific), e.g. Java, Firefox, ECMA CLR
    release: URIRef  # A project release.
    repository: URIRef  # Source code repository.
    repositoryOf: URIRef  # The project that uses a repository.
    revision: URIRef  # Revision identifier of a software release.
    screenshots: URIRef  # Web page with screenshots of project.
    shortdesc: URIRef  # Short (8 or 9 words) plain text description of a project.
    tester: URIRef  # A tester or other quality control contributor.
    translator: URIRef  # Contributor of translations to the project.
    vendor: URIRef  # Vendor organization: commercial, free or otherwise
    wiki: URIRef  # URL of Wiki for collaborative discussion of project.

    # http://www.w3.org/2000/01/rdf-schema#Class
    ArchRepository: URIRef  # GNU Arch source code repository.
    BKRepository: URIRef  # BitKeeper source code repository.
    BazaarBranch: URIRef  # Bazaar source code branch.
    CVSRepository: URIRef  # CVS source code repository.
    DarcsRepository: URIRef  # darcs source code repository.
    GitBranch: URIRef  # Git source code branch.
    GitRepository: URIRef  # Git source code repository.
    HgRepository: URIRef  # Mercurial source code repository.
    Project: URIRef  # A project.
    Repository: URIRef  # Source code repository.
    SVNRepository: URIRef  # Subversion source code repository.
    Specification: URIRef  # A specification of a system's aspects, technical or otherwise.
    Version: URIRef  # Version information of a project release.

    # http://www.w3.org/2002/07/owl#InverseFunctionalProperty
    homepage: URIRef  # URL of a project's homepage, 		associated with exactly one project.

    # Valid non-python identifiers
    _extras = [
        "anon-root",
        "bug-database",
        "developer-forum",
        "download-mirror",
        "download-page",
        "file-release",
        "mailing-list",
        "programming-language",
        "service-endpoint",
        "support-forum",
        "old-homepage",
    ]

    _NS = Namespace("http://usefulinc.com/ns/doap#")



from rdflib import Graph, Namespace, RDF, Variable

from rdflib.query import Result, ResultParser

RS = Namespace("http://www.w3.org/2001/sw/DataAccess/tests/result-set#")


class RDFResultParser(ResultParser):
    def parse(self, source, **kwargs):
        return RDFResult(source, **kwargs)


class RDFResult(Result):
    def __init__(self, source, **kwargs):

        if not isinstance(source, Graph):
            graph = Graph()
            graph.parse(source, **kwargs)
        else:
            graph = source

        rs = graph.value(predicate=RDF.type, object=RS.ResultSet)
        # there better be only one :)

        if rs is None:
            type_ = "CONSTRUCT"

            # use a new graph
            g = Graph()
            g += graph

        else:

            askAnswer = graph.value(rs, RS.boolean)

            if askAnswer is not None:
                type_ = "ASK"
            else:
                type_ = "SELECT"

        Result.__init__(self, type_)

        if type_ == "SELECT":
            self.vars = [Variable(v) for v in graph.objects(rs, RS.resultVariable)]

            self.bindings = []

            for s in graph.objects(rs, RS.solution):
                sol = {}
                for b in graph.objects(s, RS.binding):
                    sol[Variable(graph.value(b, RS.variable))] = graph.value(
                        b, RS.value
                    )
                self.bindings.append(sol)
        elif type_ == "ASK":
            self.askAnswer = askAnswer.value
            if askAnswer.value is None:
                raise Exception("Malformed boolean in ask answer!")
        elif type_ == "CONSTRUCT":
            self.graph = g
